// -----------------------------------------------------------------------------
// Copyright 2009-2012 Patrick Näf (herzbube@herzbube.ch)
//
// This file is part of the dgsmonX project.
//
// dgsmonX is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// dgsmonX is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with dgsmonX. If not, see <http://www.gnu.org/licenses/>.
// -----------------------------------------------------------------------------


// Project includes
#import "DGSMonXController.h"
#import "DGSMonXModel.h"
#import "DGSMonXPreferences.h"
#import "DGSMonXThread.h"
#import "DGSMonXServer.h"
#import "DGSMonXServerMenuItem.h"


/// @brief This category declares private methods for the DGSMonXController
/// class. 
@interface DGSMonXController()
/// @name Initialization and deallocation
//@{
- (void) dealloc;
+ (void) initialize;
//@}

/// @name NSNibAwaking protocol
//@{
- (void) awakeFromNib;
//@}

/// @name NSApplication delegate
//@{
- (void) applicationDidFinishLaunching:(NSNotification*)aNotification;
- (void) applicationWillTerminate:(NSNotification*)aNotification;
//@}

/// @name NSMenuValidation protocol
//@{
- (BOOL) validateMenuItem:(NSMenuItem*)menuItem;
//@}

/// @name Action methods for items in main menu
//@{
- (void) checkOrCancelAll:(id)sender;
- (void) serverMenuItemAction:(id)sender;
- (void) showPreferencesDialog:(id)sender;
//@}

/// @name Action methods for items in "Help" menu
//@{
- (void) showGPL:(id)sender;
- (void) showManual:(id)sender;
- (void) showReadme:(id)sender;
- (void) showChangeLog:(id)sender;
- (void) showRoadmap:(id)sender;
- (void) gotoHomepage:(id)sender;
//@}

/// @name Notification handlers
//@{
+ (void) errorConditionOccurred:(NSNotification*)aNotification;
//@}

/// @name Other methods
//@{
- (void) setMenuItem:(NSMenuItem*)item state:(bool)newState;
- (void) showTextFileInWindow:(NSString*)fileName;
- (int) monitoringThreadState;
- (void) updateServerMenuItems;
- (void) startStatusItemAnimation;
- (void) stopStatusItemAnimation;
- (void) toggleStatusItemState:(NSTimer*)timer;
- (void) setStatusItemState:(enum DGSMonXStatusItemState)state;
- (void) updateStatusItemState;
- (void) updateStatusItemAppearance;
- (void) acknowledgeStatusItemAnimation;
- (bool) itsYourTurn;
//@}
@end


@implementation DGSMonXController

// -----------------------------------------------------------------------------
/// @brief Initializes the DGSMonXController class and triggers other very
/// early initialization tasks.
///
/// As recommended by the Cocoa docs, this implementation exists because
/// DGSMonXController is the application delegate and therefore receives the
/// initialize() message very early during application launch.
///
/// @todo Maybe not early enough! Even though this is a class method, the model
/// is alloc/init'ed ***BEFORE*** this method is called. It stands to question
/// whether setting up application defaults in this method is early enough for
/// the requirements of the shared NSUserDefaultController (see topic
/// "The Shared User Defaults Controller" in the ADC companion guide
/// "User Defaults"). So far it has worked
// -----------------------------------------------------------------------------
+ (void)initialize
{
  // Guard against calls from sub-classes
  if (self != [DGSMonXController class])
    return;

  // Registering the error handler must be the very first action, so that other
  // classes may report any errors to this central error handling class
  NSNotificationCenter* center = [NSNotificationCenter defaultCenter];
  if (center)
  {
    [center addObserver:self
               selector:@selector(errorConditionOccurred:)
                   name:errorConditionOccurredNotification
                 object:nil];
  }

  // Setup application defaults
  [DGSMonXPreferences setupDefaults];
}

// -----------------------------------------------------------------------------
/// @brief Initializes an DGSMonXController object.
///
/// @note This is the designated initializer of DGSMonXController.
// -----------------------------------------------------------------------------
- (id) init
{
  // Call designated initializer of superclass (NSObject)
  self = [super init];
  if (! self)
    return nil;

  // Init members that are not initialized in MainMenu.xib
  m_statusItem = nil;
  m_monitoringThread = nil;
  m_connection = nil;
  m_monitoringThreadState = IdleThreadState;
  m_statusItemAnimationTimer = nil;
  m_statusItemState = StatusItemStateUnknown;
  m_normalStateImageBlackAndWhite = [[NSImage imageNamed:iconBlackWhiteFileName] retain];
  m_normalStateImageGrayScale = [[NSImage imageNamed:iconGrayScaleFileName] retain];
  m_itsYourTurnStateImage = [[NSImage imageNamed:iconColorFileName] retain];

  return self;
}

// -----------------------------------------------------------------------------
/// @brief Deallocates memory allocated by this DGSMonXController object.
// -----------------------------------------------------------------------------
- (void) dealloc
{
  [m_normalStateImageBlackAndWhite release];
  [m_normalStateImageGrayScale release];
  [m_itsYourTurnStateImage release];
  [super dealloc];
}

// -----------------------------------------------------------------------------
/// @brief Is called after an DGSMonXController object has been allocated and
/// initialized from MainMenu.xib
///
/// @note This is an NSNibAwaking protocol method.
// -----------------------------------------------------------------------------
- (void) awakeFromNib
{
}

// -----------------------------------------------------------------------------
/// @brief Reacts to the event that the application finishes its launching
/// process.
///
/// @a aNotification is ignored.
///
/// @note This is an NSApplication delegate method.
// -----------------------------------------------------------------------------
- (void) applicationDidFinishLaunching:(NSNotification*)aNotification
{
  // Create the status item
  // Note: It's our job to retain the item, the status bar does not do so
  NSStatusBar* bar = [NSStatusBar systemStatusBar];
  m_statusItem = [bar statusItemWithLength:NSVariableStatusItemLength];
  [m_statusItem retain];
  // Configure the item
  [m_statusItem setHighlightMode:YES];
  [m_statusItem setMenu:m_theMenu];   // status item retains the menu
  // Set up the status item's initial appearance according to user defaults.
  // The server check triggered after application launch will update the status
  // item's appearance again, but only if at least 1 server is configured. If
  // no servers are configured, updateStatusItem() is never called. If we were
  // not updating the status item here, the item would remain invisible.
  [self setStatusItemState:StatusItemStateNormal];
  [self updateStatusItemAppearance];

  // Load the initial server list from the user defaults. We must do this
  // ***BEFORE*** launching the monitoring thread because the thread will
  // immediately start its first game state check.
  [m_theModel loadServerListFromUserDefaults];

  // Prepare for communication with the monitoring thread
  NSPort* port1 = [NSPort port];
  NSPort* port2 = [NSPort port];
  m_connection = [[NSConnection alloc] initWithReceivePort:port1
                                                  sendPort:port2];
  // Retains self, creating a retain cycle. The cycle is broken when the
  // connection's invalidate() method is invoked - see
  // applicationWillTerminate:().
  [m_connection setRootObject:self];

  // Create monitoring thread. We must do this ***AFTER*** loading the model's
  // initial server list from the user defaults because the thread will
  // immediately start its first game state check.
  // Note: Because we don't know the thread instance here, the thread object
  // will invoke a callback method during its initialization to give us that
  // instance.
  NSArray* portArray = [NSArray arrayWithObjects:port2, port1, nil];
  [NSThread detachNewThreadSelector:@selector(main:)
                           toTarget:[DGSMonXThread class] withObject:portArray];

  // Requires that the model has loaded its server list
  [self updateServerMenuItems];
}

// -----------------------------------------------------------------------------
/// @brief Reacts to the event that the application is about to quit.
///
/// Sends an autorelease message to this DGSMonXController. Autoreleasing and
/// therefore deallocating this controller triggers deallocation of all the
/// other application's objects (e.g. application model).
///
/// @a aNotification is ignored.
///
/// @note This is an NSApplication delegate method.
// -----------------------------------------------------------------------------
- (void)applicationWillTerminate:(NSNotification*)aNotification
{
  // Make sure that the thread terminates. This must happen before any
  // DGSMonXServer objects that it still accesses are deallocated
  // -> After invoking this method, the object referenced by m_monitoringThread
  //    is soon going to be deallocated (as soon as the other thread has had
  //    time to do so)
  // -> Do ***NOT*** access m_monitoringThread after this !!!
  [m_monitoringThread exit];
  m_monitoringThread = nil;   // make sure that we don't do anything stupid :-)

  // Invalidates all kinds of things - see the method docs for details. The
  // important thing for us is that the invalidating process causes the
  // connection to release the retained reference to ourself (i.e. the
  // DGSMonXController instance). This breaks the retain cycle that we created
  // in our applicationDidFinishLaunching:() method.
  [m_connection invalidate];
  [m_connection release];

  // These release() messages balance the retain() messages that were sent by
  // Cocoa's nib-loading code when it loaded MainMenu.xib.
  [m_theMenu release];
  [m_thePreferences release];
  [m_theModel release];
  [m_textViewWindow release];

  // When the status item is released, it does ***NOT*** release the menu that
  // it is configured with
  // -> I consider this to be a bug!
  // -> Fix the problem by explicitly removing the menu from the status item:
  //    In this case the status item properly releases the menu.
  [m_statusItem setMenu:nil];
  [m_statusItem release];

  // The menu never gets released, because when the user selects the "quit"
  // menu item to terminate the application, the menu is temporarily visible and
  // "something" inside Cocoa retains the menu while it is visible. It appears
  // as if that "something" is then autorelease'ing the menu, but these
  // autoreleases are not executed anymore
  // -> I was unable to determine *why* this is the case. I *guess* that it
  //    must have something to do with the fact that the application is
  //    shutting down. If the menu is displayed during the normal operation of
  //    the application (i.e. not triggering a shutdown), the autoreleases
  //    occur as expected.
  // -> The effect of the skipped autoreleases is this: Because the menu is
  //    never released, the menu items it contain are also never released.
  // -> This means that our own DGSMonXServerMenuItem objects are never
  //    released, and consequently that they are never deallocated :-(
  // -> To fix this, we manually remove our DGSMonXServerMenuItem objects from
  //    the menu. This causes the menu to release the items, which eventually
  //    leads to deallocation of our DGSMonXServerMenuItem objects.
  // -> I don't know what happens to the menu, but heck! that's Cocoa's problem!
  //    The cleanup code in this method is already anal enough - after all we
  //    are in the application shutdown phase and theoretically don't need to
  //    free memory, the OS will reclaim it all anyway when the applicaton's
  //    process terminates. But hey, let's be nice and do the best we can :-)
  NSArray* serverList = [self serverList];
  NSEnumerator* enumerator = [serverList objectEnumerator];
  while (enumerator)
  {
    DGSMonXServer* server = (DGSMonXServer*)[enumerator nextObject];
    if (! server)
      break;
    [m_theMenu removeItem:[server menuItem]];
  }

  // Invoking this method invalidates the animation timer. We must do this here,
  // not in dealloc(), because dealloc() will never be called as long as the
  // animation timer is running: As long as the timer runs, it retains this
  // controller and therefore prevents deallocation.
  [self stopStatusItemAnimation];

  // This autorelease() message balances the retain() message that was sent by
  // Cocoa's nib-loading code when it loaded MainMenu.xib.
  // Note: We must *auto*release, i.e. we can't do a hard release, because if
  // we did we would immediately deallocate this instance, even though code
  // execution must still return from this instance method.
  /// @todo Test whether this assertion is true, or whether it is possible to
  /// do a hard release *IF* this is the last statement in this method and no
  /// more code besides the return from this function needs to be executed.
  [self autorelease];
}

// -----------------------------------------------------------------------------
/// @brief Makes the monitoring thread object known to this controller.
///
/// @note This method is invoked by the monitoring thread.
// -----------------------------------------------------------------------------
- (void) setMonitoringThread:(id)anObject
{
  // anObject is actually an instance of NSDistantObject, but since it behaves
  // exactly like a DGSMonXThread instance (we prove this with isKindOfClass()),
  // we can make a cast and treat it as such.
  // Note: This may not be good practice (and it would almost certainly cause
  // trouble to do the same in the C++ world), but we do it anyway, because
  // "it works (tm)" and we are reassured by the isKindOfClass() introspection
  // check. God bless dynamic typing!!!
  if ([anObject isKindOfClass:[DGSMonXThread class]])
    m_monitoringThread = (DGSMonXThread*)anObject;   // don't retain, object will remain valid as long as m_connection is not invalidated
  else
    [[NSNotificationCenter defaultCenter] postNotificationName:errorConditionOccurredNotification
                                                        object:@"An internal error has occurred: Object is not an instance of DGSMonXThread."];
}

// -----------------------------------------------------------------------------
/// @brief Returns the list of DGSMonXServer objects currently known to the
/// application.
///
/// @note This method gets its data from the application model. This controller
/// therefore effectively acts as a proxy for the application model. This
/// approach was implemented to reduce the coupling of DGSMonXThread.
///
/// @note This method is invoked by the monitoring thread.
// -----------------------------------------------------------------------------
- (NSArray*) serverList
{
  if (m_theModel)
    return [m_theModel serverList];
  else
    return nil;
}

// -----------------------------------------------------------------------------
/// @brief Returns true if @a menuItem should be enabled, false if it should be
/// disabled.
///
/// @note This is an NSMenuValidation protocol method.
///
/// @note This method also handles the user's acknowledgment (as described in
/// the MANUAL) of the status item animation. The result of this handling is
/// that the status item is put back into the normal state. It would be
/// preferrable if we could use the NSMenu delegate method menuWillOpen:(),
/// unfortunately this method is available only since Mac OS X 10.5.
// -----------------------------------------------------------------------------
- (BOOL) validateMenuItem:(NSMenuItem*)menuItem
{
  if (menuItem == m_checkOrCancelAllMenuItem)
  {
    // Special handling for the user's acknowledgment of the status item
    // animation (see method documentation). From a conceptual point of
    // view, this handling has nothing to do with the "check/cancel all" menu
    // item. The implementation exists here merely because we don't have a
    // better place for it. We simply assume that the NSMenuValidation protocol
    // is run exactly once for the "check/cancel all" menu item, after the user
    // has clicked the application's status item.
    [self acknowledgeStatusItemAnimation];

    // Regular handling for m_checkOrCancelAllMenuItem starts here
    NSString* title;
    BOOL enable;
    switch ([self monitoringThreadState])
    {
      case IdleThreadState:
        title = @"Check all";
        enable = YES;
        break;
      case CheckingThreadState:
        title = @"Cancel all checks";
        enable = YES;
        break;
      default:
        title = @"<Internal error, no action possible>";
        enable = NO;
        break;
    }
    /// @todo Should also disable menu item if we have do have servers, but
    /// monitoring is disabled for all of them.
    if ([[self serverList] count] == 0)
      enable = NO;
    [m_checkOrCancelAllMenuItem setTitle:title];
    return enable;
  }
  else if ([menuItem isKindOfClass:[DGSMonXServerMenuItem class]])
  {
    /// @todo Thread safety issue here! The update() method is executed in the
    /// context of the main thread (read access of server attributes), but at
    /// the same time a DGSMonXServerCheck might report back its results in the
    /// context of the monitoring thread (write access of the same server
    /// attributes). We should introduce a synchronization mechanism somewhere
    /// in DGSMonXServer. Write this down somewhere in a central document,
    /// because so far this has been the only synchronization issue that we had
    /// to take care of.
    DGSMonXServerMenuItem* serverMenuItem = (DGSMonXServerMenuItem*)menuItem;
    [serverMenuItem update];
    return [[serverMenuItem server] monitored];
  }
  else
    return YES;
}

// -----------------------------------------------------------------------------
/// @brief Either triggers a game state check for a single server, or cancels
/// a check that is in progress for the server.
///
/// Which one of the actions is performed depends on what the monitoring thread
/// is doing at the moment:
/// - If the thread is currently performing a game state check for the server,
///   this action cancels that check.
/// - If the thread is currently not performing a check, this action triggers a
///   new game state check.
///
/// @note This is an action method.
// -----------------------------------------------------------------------------
- (void) serverMenuItemAction:(id)sender
{
  DGSMonXServer* server = [(DGSMonXServerMenuItem*)sender server];
  if (! server)
  {
    assert(0);
    return;
  }

  NSUserDefaults* userDefaults = [NSUserDefaults standardUserDefaults];
  enum DGSMonXMenuItemClickAction menuItemClickAction = [userDefaults integerForKey:menuItemClickActionKey];
  switch (menuItemClickAction)
  {
    case MenuItemClickActionOpenBrowser:
    {
      NSString* baseURL = [server baseURL];
      NSString* pageURL = [baseURL stringByAppendingString:statusPage];
      [[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:pageURL]];
      break;
    }
    case MenuItemClickActionGameStateCheck:
    default:
    {
      /// @todo Thread safety issue here! The checkState() method should use a
      /// mutex to protect the state from being changed by a DGSMonXServerCheck
      /// right now.
      switch ([server checkState])
      {
        case InProgressCheckState:
          [m_monitoringThread cancelSingle:server];
          break;
        case NonePerformedCheckState:
        case DidFinishCheckState:
        case WasCancelledCheckState:
          [m_monitoringThread checkSingle:server];
          break;
        default:
          // do nothing; this should not happen anyway...
          break;
      }
      break;
    }
  }
}

// -----------------------------------------------------------------------------
/// @brief Either triggers a series of game state checks for all servers, or
/// cancels all checks in progress.
///
/// Which one of the actions is performed depends on what the monitoring thread
/// is doing at the moment:
/// - If the thread is currently performing game state checks, this action
///   cancels those checks.
/// - If the thread is currently idle, this action triggers a new series of
///   game state checks.
///
/// @note This is an action method.
// -----------------------------------------------------------------------------
- (void) checkOrCancelAll:(id)sender
{
  switch ([self monitoringThreadState])
  {
    case IdleThreadState:
      [m_monitoringThread checkAll];
      break;
    case CheckingThreadState:
      [m_monitoringThread cancelAll];
      break;
    default:
      // do nothing; this should not happen anyway...
      break;
  }
}

// -----------------------------------------------------------------------------
/// @brief Displays the user preferences dialog.
///
/// @note This is an action method.
// -----------------------------------------------------------------------------
- (void) showPreferencesDialog:(id)sender
{
  /// @todo This will block if the monitoring thread is currently checking,
  /// i.e. if the run loop is busy. If we send the message asynchronously, this
  /// will not block, but the current check will not be interrupted,
  /// either - the asynchronous message will simply arrive delayed, i.e. when
  /// the run loop is free again.
  [m_monitoringThread suspendMonitoring];

  // Remember current values so that we can check after the dialog closes
  // whether the user has changed them
  NSUserDefaults* userDefaults = [NSUserDefaults standardUserDefaults];
  int unit = [userDefaults integerForKey:monitoringIntervalUnitKey];
  int amount = [userDefaults integerForKey:monitoringIntervalAmountKey];

  // The dialog is modal
  // -> the method call does not return until the dialog is closed
  [m_thePreferences showPreferencesDialog];

  // Write changes to the server list to user defaults
  [m_theModel writeServerListToUserDefaults];

  // Add menu items for *NEW* servers and remove menu items for *DELETED*
  // servers.
  [self updateServerMenuItems];

  // Notify monitoring thread that the interval has changed. In reaction there
  // will be some rescheduling of checks.
  if (unit != [userDefaults integerForKey:monitoringIntervalUnitKey] ||
      amount != [userDefaults integerForKey:monitoringIntervalAmountKey])
    [m_monitoringThread monitoringIntervalHasChanged];

  // Tell the monitoring thread that it is again allowed to do its duties
  [m_monitoringThread resumeMonitoring];

  // Update status item state in case server's monitoring state was
  // enabled/disabled
  [self updateStatusItemState];

  // Update status item appearance to reflect the current user defaults (in case
  // the user did change any of her preferences)
  [self updateStatusItemAppearance];
}

// -----------------------------------------------------------------------------
/// @brief Displays a window that contains the GPL license information (i.e.
/// the COPYING file).
///
/// @note This is an action method.
// -----------------------------------------------------------------------------
- (void) showGPL:(id)sender
{
  [self showTextFileInWindow:gnuGPLFileName];
}

// -----------------------------------------------------------------------------
/// @brief Displays a window that contains the MANUAL file.
///
/// @note This is an action method.
// -----------------------------------------------------------------------------
- (void) showManual:(id)sender
{
  [self showTextFileInWindow:manualFileName];
}

// -----------------------------------------------------------------------------
/// @brief Displays a window that contains the README file.
///
/// @note This is an action method.
// -----------------------------------------------------------------------------
- (void) showReadme:(id)sender
{
  [self showTextFileInWindow:readMeFileName];
}

// -----------------------------------------------------------------------------
/// @brief Displays a window that contains the ChangeLog file.
///
/// @note This is an action method.
// -----------------------------------------------------------------------------
- (void) showChangeLog:(id)sender
{
  [self showTextFileInWindow:changeLogFileName];
}

// -----------------------------------------------------------------------------
/// @brief Displays a window that contains the Roadmap file.
///
/// @note This is an action method.
// -----------------------------------------------------------------------------
- (void) showRoadmap:(id)sender
{
  [self showTextFileInWindow:roadmapFileName];
}

// -----------------------------------------------------------------------------
/// @brief Opens the dgsmonX homepage URL in the default browser.
///
/// @note This is an action method.
// -----------------------------------------------------------------------------
- (void) gotoHomepage:(id)sender
{
  [[NSWorkspace sharedWorkspace] openURL:[NSURL URLWithString:homePageURL]];
}

// -----------------------------------------------------------------------------
/// @brief Performs application-wide error handling.
///
/// This method is called by the NSNotificationCenter when a severe error
/// condition is detected. Trivial errors should be handled by the code that
/// detects the error.
///
/// If a method detects a severe error condition it should take the
/// appropriate steps for a first-level reaction to the error (usually it
/// just terminates/aborts the operation it was supposed to perform in a
/// clean way). Then it it should compose a text (ideally one or more
/// whole sentences) describing the error condition and post it together
/// with an NSNotification to the NSNotificationCenter.
///
/// This method then reacts to the notification by displaying an alert dialog
/// to the user, informing her about the problem and including the error
/// description. The dialog offers the user to terminate the application
/// (the default button), or to ignore the error.
// -----------------------------------------------------------------------------
+ (void) errorConditionOccurred:(NSNotification*)aNotification
{
  // Make dgsmonX the active app so that the window we are going to display
  // is guaranteed to appear in front of other applications' windows.
  // Note: This is only necessary because dgsmonX is faceless and cannot become
  // the active application through normal user interaction.
  NSApplication* theApp = [NSApplication sharedApplication];
  [theApp activateIgnoringOtherApps:YES];

  NSAlert* alert = [NSAlert alertWithMessageText:@"dgsmonX has encountered a critical error!"
                                   defaultButton:@"Terminate application"
                                 alternateButton:@"Ignore error & continue"
                                     otherButton:nil
                       informativeTextWithFormat:(NSString*)[aNotification object]];
  [alert setAlertStyle:NSCriticalAlertStyle];
  int buttonClicked = [alert runModal];
  // Terminate the application if default button was clicked
  // Note: the documentation for the "runModal" method says that the method's
  // return values are the "new" (=more modern) values NSAlertFirstButtonReturn,
  // NSAlertSecondButtonReturn, etc. This does ***NOT*** apply to our case
  // because we have created the alert using "alertWithMessageText...", and this
  // method's documentation clearly states that the "old" return values
  // NSAlertDefaultReturn, NSAlertAlternateReturn, etc. are used.
  if (NSAlertDefaultReturn == buttonClicked)
    [[NSApplication sharedApplication] terminate:self];
}

// -----------------------------------------------------------------------------
/// @brief Sets the enabled state of @a item to the new state @a newState.
///
/// @note This is an internal helper method.
// -----------------------------------------------------------------------------
- (void) setMenuItem:(NSMenuItem*)item state:(bool)newState
{
  if (newState)
    [item setState:NSOnState];
  else
    [item setState:NSOffState];
}

// -----------------------------------------------------------------------------
/// @brief Shows the content of file @a fileName in a separate window.
///
/// The requested file must be a UTF-8 encoded text file and it must be part of
/// the application bundle.
///
/// @note This is an internal helper method.
// -----------------------------------------------------------------------------
- (void) showTextFileInWindow:(NSString*)fileName
{
  // Make dgsmonX the active app so that the window we are going to display
  // is guaranteed to appear in front of other applications' windows.
  // Note: This is only necessary because dgsmonX is faceless and cannot become
  // the active application through normal user interaction.
  NSApplication* theApp = [NSApplication sharedApplication];
  [theApp activateIgnoringOtherApps:YES];

  // Try to get the file's content
  NSString* textFileContent = nil;
  NSBundle* mainBundle = [NSBundle mainBundle];
  if (mainBundle)
  {
    NSString* textFilePath = [mainBundle pathForResource:fileName ofType:nil];
    if (textFilePath)
    {
      // Mac OS X 10.4 API
      textFileContent = [NSString stringWithContentsOfFile:textFilePath
                                                  encoding:NSUTF8StringEncoding
                                                     error:nil];
    }
  }

  if (textFileContent)
  {
    [m_textView setString:textFileContent];
    [m_textViewWindow setTitle:fileName];
    [m_textViewWindow makeKeyAndOrderFront:self];
    // No need to call makeMainWindow()
  }
  else
  {
    NSAlert* alert = [NSAlert alertWithMessageText:@"Could not open file"
                                     defaultButton:nil
                                   alternateButton:nil
                                       otherButton:nil
                         informativeTextWithFormat:fileName];
    [alert setAlertStyle:NSInformationalAlertStyle];
    [alert runModal];
  }
}

// -----------------------------------------------------------------------------
/// @brief Returns the current state of the monitoring thread.
///
/// The return value of this method is a member of the enumeration
/// #DGSMonXThreadState.
// -----------------------------------------------------------------------------
- (int) monitoringThreadState
{
  return m_monitoringThreadState;
}

// -----------------------------------------------------------------------------
/// @brief Sets the state of the monitoring thread to @a state.
///
/// @a state must be a member of the enumeration #DGSMonXThreadState.
///
/// @note This method is invoked by the monitoring thread.
// -----------------------------------------------------------------------------
- (void) setMonitoringThreadState:(int)state
{
  m_monitoringThreadState = state;
}

// -----------------------------------------------------------------------------
/// @brief Reacts to a game state check having finished by updating the
/// appearance of the application's status item.
///
/// @note This method is invoked by the monitoring thread.
// -----------------------------------------------------------------------------
- (void) gameStateCheckDidFinish
{
  NSUserDefaults* userDefaults = [NSUserDefaults standardUserDefaults];
  enum DGSMonXNotificationStyle notificationStyle = [userDefaults integerForKey:notifyStyleKey];
  if (NotificationStyleAnimation == notificationStyle)
  {
    bool itsYourTurn = [self itsYourTurn];
    if (itsYourTurn)
      [self startStatusItemAnimation];
    else
      [self stopStatusItemAnimation];
  }

  [self updateStatusItemState];
  [self updateStatusItemAppearance];
}

// -----------------------------------------------------------------------------
/// @brief Adds or removes menu items that represent DGSMonXServer objects from
/// the menu associated with the application's status item.
///
/// Menu items that are present in the menu, but whose DGSMonXServer object no
/// longer exists in the application model, are removed from the menu. It is
/// assumed that the DGSMonXServer objects have been deallocated either by
/// removing the corresponding servers in the Preferences dialog, or by the
/// application shutdown procedure.
///
/// Menu items that are not yet present in the menu, but whose DGSMonXServer
/// object still exists in the application model, are added to the menu. It is
/// assumed that the DGSMonXServer objects have been created either by adding
/// the corresponding servers in the Preferences dialog, or by the application
/// startup procedure.
///
/// @note This is an internal helper method.
// -----------------------------------------------------------------------------
- (void) updateServerMenuItems;
{
#if (MAC_OS_X_VERSION_MAX_ALLOWED <= MAC_OS_X_VERSION_10_4)
  int menuIndex;
  unsigned arrayIndex;
#else
  NSInteger menuIndex;
  NSUInteger arrayIndex;
#endif
  NSArray* serverList = [self serverList];

  // Pass 1: Remove menu items whose servers no longer exist
  menuIndex = 0;
  while (true)
  {
    NSMenuItem* menuItem = [m_theMenu itemAtIndex:menuIndex];
    if (! [menuItem isKindOfClass:[DGSMonXServerMenuItem class]])
      break;
    DGSMonXServer* server = [(DGSMonXServerMenuItem*)menuItem server];
    if (! [serverList containsObject:server])
      [m_theMenu removeItem:menuItem];
    else
      ++menuIndex;
  }

  // Pass 2: Add menu items that are not yet present
  arrayIndex = 0;
  NSEnumerator* enumerator = [serverList objectEnumerator];
  while (enumerator)
  {
    DGSMonXServer* server = (DGSMonXServer*)[enumerator nextObject];
    if (! server)
      break;
    DGSMonXServerMenuItem* menuItem = [server menuItem];
    if (! [menuItem menu])
    {
      [m_theMenu insertItem:menuItem atIndex:arrayIndex];
      // We must specify self as the target *AND* a valid action method. If we
      // don't do this, our own validateMenuItem:() is not invoked when the
      // menu item is about to be displayed.
      [menuItem setTarget:self];
      [menuItem setAction:@selector(serverMenuItemAction:)];
    }
    ++arrayIndex;
  }
}

// -----------------------------------------------------------------------------
/// @brief Starts the status item animation.
///
/// Calling this method has no effect if the status item animation is already
/// started.
///
/// @note This is an internal helper method.
// -----------------------------------------------------------------------------
- (void) startStatusItemAnimation
{
  // Assume that the animation is started if we have a timer
  if (m_statusItemAnimationTimer)
    return;

  NSUserDefaults* userDefaults = [NSUserDefaults standardUserDefaults];
  float interval = [userDefaults floatForKey:notifyAnimateStatusItemIntervalKey];

  // Timer is added automatically to run loop. We don't have to retain because
  // the run loop retains the timer for us.
  m_statusItemAnimationTimer = [NSTimer scheduledTimerWithTimeInterval:interval
                                                                target:self
                                                              selector:@selector(toggleStatusItemState:)
                                                              userInfo:nil
                                                               repeats:YES];
}

// -----------------------------------------------------------------------------
/// @brief Stops the status item animation.
///
/// Calling this method has no effect if the status item animation is already
/// stopped.
///
/// @note This is an internal helper method.
// -----------------------------------------------------------------------------
- (void) stopStatusItemAnimation
{
  if (! m_statusItemAnimationTimer)
    return;
  [m_statusItemAnimationTimer invalidate];
  m_statusItemAnimationTimer = nil;
}

// -----------------------------------------------------------------------------
/// @brief Toggles the state of the application's status item.
///
/// If the status item is currently in its "normal" state, this method switches
/// it to the "it's your turn" state. And vice versa.
///
/// @a timer is ignored. The parameter exists solely because the NSTimer
/// interface requires a method signature that has the NSTimer object as its
/// sole argument.
///
/// @note This is an internal helper method.
// -----------------------------------------------------------------------------
- (void) toggleStatusItemState:(NSTimer*)timer
{
  switch (m_statusItemState)
  {
    case StatusItemStateNormal:
    {
      [self setStatusItemState:StatusItemStateItsYourTurn];
      break;
    }
    case StatusItemStateItsYourTurn:
    {
      [self setStatusItemState:StatusItemStateNormal];
      break;
    }
    default:
    {
      assert(0);
      break;
    }
  }
  [self updateStatusItemAppearance];
}

// -----------------------------------------------------------------------------
/// @brief Sets the state of the application's status item to @a state.
///
/// @note This is an internal helper method.
// -----------------------------------------------------------------------------
- (void) setStatusItemState:(enum DGSMonXStatusItemState)state
{
  if (m_statusItemState == state)
    return;
  m_statusItemState = state;
}

// -----------------------------------------------------------------------------
/// @brief Updates the state of the application's status item to reflect the
/// current global "it's your turn" state.
///
/// @note This is an internal helper method.
// -----------------------------------------------------------------------------
- (void) updateStatusItemState
{
  bool itsYourTurn = [self itsYourTurn];
  if (itsYourTurn)
    [self setStatusItemState:StatusItemStateItsYourTurn];
  else
    [self setStatusItemState:StatusItemStateNormal];
}

// -----------------------------------------------------------------------------
/// @brief Updates the appearance of the application's status item to reflect
/// the status item state (normal/it's your turn), combined with the current
/// user defaults.
///
/// @note This is an internal helper method.
// -----------------------------------------------------------------------------
- (void) updateStatusItemAppearance
{
  NSUserDefaults* userDefaults = [NSUserDefaults standardUserDefaults];
  int statusItemStyle = [userDefaults integerForKey:statusItemStyleKey];
  NSString* statusItemNormalText = [userDefaults stringForKey:statusItemNormalTextKey];
  NSString* statusItemItsYourTurnText = [userDefaults stringForKey:statusItemItsYourTurnTextKey];
  int statusItemIconStyle = [userDefaults integerForKey:statusItemIconStyleKey];
  BOOL statusItemDisplayNumberOfGames = [userDefaults boolForKey:statusItemDisplayNumberOfGamesKey];

  bool statusItemDisplayIcon = false;
  bool statusItemDisplayText = false;
  switch (statusItemStyle)
  {
    case StatusItemStyleTextOnly:
    {
      statusItemDisplayText = true;
      break;
    }
    case StatusItemStyleBothIconAndText:
    {
      statusItemDisplayText = true;
      statusItemDisplayIcon = true;
      break;
    }
    case StatusItemStyleIconOnly:
    default:
    {
      statusItemDisplayIcon = true;
      break;
    }
  }

  int numberOfGames = 0;
  if (statusItemDisplayNumberOfGames)
  {
    NSArray* serverList = [self serverList];
    NSEnumerator* enumerator = [serverList objectEnumerator];
    while (enumerator)
    {
      DGSMonXServer* server = (DGSMonXServer*)[enumerator nextObject];
      if (! server)
        break;
      if (! [server monitored])
        continue;
      if ([server itsYourTurn])
        numberOfGames += [server numberOfGames];
    }
  }

  NSImage* statusItemImage = nil;
  NSString* statusItemText = nil;
  switch (m_statusItemState)
  {
    case StatusItemStateNormal:
    {
      if (statusItemDisplayIcon)
      {
        switch (statusItemIconStyle)
        {
          case StatusItemIconStyleGrayscale:
          {
            statusItemImage = m_normalStateImageGrayScale;
            break;
          }
          case StatusItemIconStyleBlackAndWhite:
          default:
          {
            statusItemImage = m_normalStateImageBlackAndWhite;
            break;
          }
        }
      }
      if (statusItemDisplayText)
        statusItemText = statusItemNormalText;
      if (statusItemDisplayNumberOfGames)
      {
        // ignore "number of games" if the status item is in the normal state
        // because this would always show 0, and there's simply no point in
        // displaying information that the user already knows about
      }
      break;
    }
    case StatusItemStateItsYourTurn:
    {
      if (statusItemDisplayIcon)
        statusItemImage = m_itsYourTurnStateImage;
      if (statusItemDisplayText && statusItemDisplayNumberOfGames)
        statusItemText = [NSString stringWithFormat:@"%@ (%d)", statusItemItsYourTurnText, numberOfGames];
      else if (statusItemDisplayText)
        statusItemText = statusItemItsYourTurnText;
      else if (statusItemDisplayNumberOfGames)
        statusItemText = [NSString stringWithFormat:@"%d", numberOfGames];
      break;
    }
    default:
    {
      assert(0);
      break;
    }
  }

  [m_statusItem setImage:statusItemImage];
  [m_statusItem setTitle:statusItemText];
}

// -----------------------------------------------------------------------------
/// @brief Handles the user's acknowledgment (as described in the MANUAL) of
/// the status item animation.
///
/// The result of this handling is that the status item animation stops (if it
/// is currently running). The status item is then put into the state reported
/// by itsYourTurn(). Usually this will be #StatusItemStateItsYourTurn.
///
/// @note This is an internal helper method.
// -----------------------------------------------------------------------------
- (void) acknowledgeStatusItemAnimation
{
  [self stopStatusItemAnimation];
  [self updateStatusItemState];
  [self updateStatusItemAppearance];
}

// -----------------------------------------------------------------------------
/// @brief Returns the global "it's your turn" state, i.e. returns true if at
/// least one server reports that "it's your turn". Returns false if all servers
/// report that "it's NOT your turn".
///
/// @note This is an internal helper method.
// -----------------------------------------------------------------------------
- (bool) itsYourTurn
{
  bool itsYourTurn = false;
  NSArray* serverList = [self serverList];
  NSEnumerator* enumerator = [serverList objectEnumerator];
  while (enumerator)
  {
    DGSMonXServer* server = (DGSMonXServer*)[enumerator nextObject];
    if (! server)
      break;
    if (! [server monitored])
      continue;
    if ([server itsYourTurn])
    {
      itsYourTurn = true;
      break;
    }
  }
  return itsYourTurn;
}


@end
