// -----------------------------------------------------------------------------
// Copyright 2009-2012 Patrick Näf (herzbube@herzbube.ch)
//
// This file is part of the dgsmonX project.
//
// dgsmonX is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// dgsmonX is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with dgsmonX. If not, see <http://www.gnu.org/licenses/>.
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
// The following line is required so that doxygen documents the global items
// in this file.
/// @file
///
/// @brief This file contains declarations of global items that are frequently
/// used in all classes of the project. This file is included via the project's
/// precompiled prefix header dgsmonX_Prefix.pch.
// -----------------------------------------------------------------------------


// -----------------------------------------------------------------------------
/// @brief The style of the status item that represents DGSMonX in the status
/// line.
// -----------------------------------------------------------------------------
enum DGSMonXStatusItemStyle
{
  StatusItemStyleIconOnly,
  StatusItemStyleTextOnly,
  StatusItemStyleBothIconAndText
};

// -----------------------------------------------------------------------------
/// @brief The icon used for the status item that represents DGSMonX in the
/// status line.
// -----------------------------------------------------------------------------
enum DGSMonXStatusItemIconStyle
{
  StatusItemIconStyleBlackAndWhite,
  StatusItemIconStyleGrayscale
};

// -----------------------------------------------------------------------------
/// @brief The state of the status item that represents DGSMonX in the status
/// line.
// -----------------------------------------------------------------------------
enum DGSMonXStatusItemState
{
  StatusItemStateUnknown,
  StatusItemStateNormal,
  StatusItemStateItsYourTurn
};

// -----------------------------------------------------------------------------
/// @brief The style used to notify the user when "it's your turn"
// -----------------------------------------------------------------------------
enum DGSMonXNotificationStyle
{
  NotificationStyleAnimation,
  NotificationStyleStatic
};

// -----------------------------------------------------------------------------
/// @brief What happens if the user clicks a menu item representing a server or
/// game.
// -----------------------------------------------------------------------------
enum DGSMonXMenuItemClickAction
{
  MenuItemClickActionGameStateCheck,
  MenuItemClickActionOpenBrowser
};

// -----------------------------------------------------------------------------
/// @brief Protocols that a DGSMonXServer can use.
// -----------------------------------------------------------------------------
enum DGSMonXServerProtocol
{
  HttpProtocol,
  HttpsProtocol
};

// -----------------------------------------------------------------------------
/// @brief Time units known to dgsmonX.
// -----------------------------------------------------------------------------
enum DGSMonXTimeUnit
{
  HoursTimeUnit,
  MinutesTimeUnit,
  SecondsTimeUnit
};

// -----------------------------------------------------------------------------
/// @brief States of the monitoring thread.
// -----------------------------------------------------------------------------
enum DGSMonXThreadState
{
  IdleThreadState,      ///< @brief No game state checks in progress.
  CheckingThreadState   ///< @brief Game state checks are in progress.
};

// -----------------------------------------------------------------------------
/// @brief States of a DGSMonXServer relating to game state checks.
// -----------------------------------------------------------------------------
enum DGSMonXServerCheckState
{
  NonePerformedCheckState,   ///< @brief No game state check has been performed for the server (hence no results are available).
  InProgressCheckState,      ///< @brief A game state check is currently in progress for the server.
  DidFinishCheckState,       ///< @brief The last game state check performed on the server has finished (results are available).
  WasCancelledCheckState     ///< @brief The last game state check performed on the server was cancelled (hence no results are available).
};

// -----------------------------------------------------------------------------
/// @name Notifications
// -----------------------------------------------------------------------------
//@{
/// @brief Any method that encounters an error condition that it cannot handle
/// properly may post this notification. The user will be presented with a
/// dialog that offers terminating the application, or simply ignoring the
/// error.
static NSString* errorConditionOccurredNotification = @"ErrorConditionOccurred";
//@}

// -----------------------------------------------------------------------------
/// @name Constants related to user defaults
// -----------------------------------------------------------------------------
//@{
static NSString* userDefaultsVersionRegistrationDomainKey = @"UserDefaultsVersionRegistrationDomain";
static NSString* userDefaultsVersionApplicationDomainKey = @"UserDefaultsVersionApplicationDomain";
static NSString* monitoringIntervalUnitKey = @"MonitoringIntervalUnit";
static NSString* monitoringIntervalAmountKey = @"MonitoringIntervalAmount";
static NSString* statusItemStyleKey = @"StatusItemStyle";
static NSString* statusItemNormalTextKey = @"StatusItemNormalText";
static NSString* statusItemItsYourTurnTextKey = @"StatusItemItsYourTurnText";
static NSString* statusItemIconStyleKey = @"StatusItemIconStyle";
static NSString* statusItemDisplayNumberOfGamesKey = @"StatusItemDisplayNumberOfGames";
static NSString* menuListStyleKey = @"MenuListStyle";
static NSString* menuItemDisplayOpponentKey = @"MenuItemDisplayOpponent";
static NSString* menuItemClickActionKey = @"MenuItemClickAction";
static NSString* notifyStyleKey = @"NotifyStyle";
static NSString* notifyAnimateStatusItemIntervalKey = @"NotifyAnimateStatusItemInterval";
static NSString* notifyPlaySoundKey = @"NotifyPlaySound";
static NSString* serverListKey = @"ServerList";
static NSString* serverMonitoredKey = @"ServerMonitored";
static NSString* serverNameKey = @"ServerName";
static NSString* serverProtocolKey = @"ServerProtocol";
static NSString* serverURLKey = @"ServerURL";
static NSString* userIDKey = @"UserID";
static NSString* userPasswordKey = @"UserPassword";
//@}

// -----------------------------------------------------------------------------
/// @name Resource file names
// -----------------------------------------------------------------------------
//@{
static NSString* registrationDomainDefaultsFileName = @"RegistrationDomainDefaults.plist";
// Text file names
static NSString* gnuGPLFileName = @"COPYING";
static NSString* manualFileName = @"MANUAL";
static NSString* readMeFileName = @"README";
static NSString* changeLogFileName = @"ChangeLog";
static NSString* roadmapFileName = @"Roadmap";
static NSString* iconColorFileName = @"dgs-color.png";
static NSString* iconBlackWhiteFileName = @"dgs-bw.png";
static NSString* iconGrayScaleFileName = @"dgs-gray.png";
//@}

// -----------------------------------------------------------------------------
/// @name HTTP related constants
// -----------------------------------------------------------------------------
//@{
static NSString* loginPage = @"login.php";
static NSString* statusPage = @"status.php";
static NSString* quickStatusPage = @"quick_status.php";
static int quickStatusPageVersion = 2;
static NSString* errorPage = @"error.php";
static NSString* handleCookie = @"cookie_handle";   // this cookie contains the user ID
static NSString* sessionCodeCookie = @"cookie_sessioncode";
//@}

// -----------------------------------------------------------------------------
/// @name NSError related constants
// -----------------------------------------------------------------------------
//@{
static NSString* errorDomain = @"dgsmonX";  ///< @brief The error domain to be used for constructing NSError objects
/// @brief Enumerates error codes that can be used to construct NSError objects.
enum DGSMonXErrorCode
{
  ErrorCodeStatusItemTextIsEmpty,
  ErrorCodeStatusItemTextIsWhitespace
};
//@}

// -----------------------------------------------------------------------------
/// @name Other constants
// -----------------------------------------------------------------------------
//@{
static NSString* newServerName = @"New DGS server";
static NSString* newServerURL = @"www.dragongoserver.net/";
static NSString* homePageURL = @"http://www.herzbube.ch/dgsmonx";
//@}
