// -----------------------------------------------------------------------------
// Copyright 2009-2012 Patrick Näf (herzbube@herzbube.ch)
//
// This file is part of the dgsmonX project.
//
// dgsmonX is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// dgsmonX is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with dgsmonX. If not, see <http://www.gnu.org/licenses/>.
// -----------------------------------------------------------------------------


// Project includes
#import "DGSMonXServerQuickStatus.h"

// Forward declarations
@class DGSMonXServerCheck;
@class DGSMonXServer;


// -----------------------------------------------------------------------------
/// @brief The DGSMonXServerCheckDelegate protocol must be implemented by the
/// delegate of DGSMonXServerCheck.
///
/// Once the delegate receives the gameStateCheckDidFinish:() message, it can
/// start querying the DGSMonXServerCheck object for the results of the check.
///
/// @note The callback message will be sent in the context of the thread that
/// initiated the game state check. For the callback to work correctly the
/// initiating thread’s run loop must be operating in the default run loop mode.
/// This restriction exists because DGSMonXServerCheck exposes a restriction of
/// the same nature in the underlying NSURLConnection. DGSMonXServerCheck does
/// @b not expose NSURLConnection facilities to circumvent the restriction,
/// because this would certainly be excessive for such a simple application as
/// dgsmonX.
// -----------------------------------------------------------------------------
@protocol DGSMonXServerCheckDelegate
- (void) gameStateCheckDidFinish:(DGSMonXServerCheck*)serverCheck;
@end


// -----------------------------------------------------------------------------
/// @brief The DGSMonXServerCheck class performs a game state check on a
/// Dragon Go Server installation.
///
/// DGSMonXServerCheck starts the game state check immediately upon receiving
/// the initWithServer:() message. The DGSMonXServer object specified represents
/// the Dragon Go Server on which the check should be performed. The check runs
/// asynchronously. When the check is done, DGSMonXServerCheck stores the
/// results in the DGSMonXServer object, then calls back the delegate that was
/// specified on initialization, sending it a message that is defined by the
/// DGSMonXServerCheckDelegate informal protocol. Upon receiving the message,
/// the delegate can start querying the DGSMonXServer object for the results of
/// the game state check.
///
/// @note The callback message will be sent in the context of the thread that
/// initiated the game state check. For the callback to work correctly the
/// initiating thread’s run loop must be operating in the default run loop mode.
///
/// The game state check can be cancelled any time before the delegate receives
/// the gameStateCheckDidFinish:() message by sending the DGSMonXServerCheck a
/// cancel() message. In this case the delegate will @b not receive the
/// gameStateCheckDidFinish:() message.
///
///
/// @par Game state check procedure
///
/// The game state check procedure works as follows:
/// - Check whether NSHTTPCookieStorage has DGS cookies for the DGS server's
///   login.php page; if so, they are temporarily removed so as not to interfere
///   with our checking procedure
///   - NSHTTPCookieStorage might have cookies because the user has visited the
///     DGS server in Safari, or another application that uses Cocoa's shared
///     cookie storage system
///   - At the time that dgsmonX was written, login.php (which we are going
///     to access in the next step) ignored any cookies that were present. We
///     still implement temporary cookie removal because we want to guard
///     against future changes of the behaviour of login.php
///   - We need cookie handling anyway because login.php ***overwrites***
///     cookies. Thus, if we do not handle cookies our checking procedure will
///     have side-effects on Safari & Co. We do our best to prevent these
///     side-effects by restoring all original DGS cookies after our checking
///     procedure finishes
///   - There is still a chance for conflict between dgsmonX and Safari & Co: If
///     the user happens to access the DGS server in Safari & Co while our game
///     state check is in progress, she will have either no cookies, or those
///     cookies that are created by the activities of the game state check
///     check. At the same time, if the user attempts a login on the DGS server
///     in Safari & Co, while our game state check is in progress, the cookies
///     generated by that login might influence our game state check
///   - DGS cookies are recognized by hard-coded names
/// - Create an NSURLRequest that points to the DGS server's login.php
///   - User ID and password are used to construct the URL, which means they
///     may be visible to network sniffers (if HTTP is used), and they may be
///     written to the HTTP server's log files on the DGS server machine (both
///     for HTTP and HTTPS)
/// - The response to the NSURLRequest is examined in order to detect whether
///   the login attempt succeeded or failed
///   - If the DGS server redirects us to its error.php, we assume that the
///     login attempt has failed. The check is terminated. We expect that the
///     URL to error.php contains a query string with a reason for the error
///     (e.g. "error.php?err=wrong_password"). We extract the reason and use it
///     to detail the game state check result.
///   - If the DGS server redirects us to some other page (i.e. not error.php),
///     we assume that the login attempt was successful and continue the check
/// - Create an NSURLRequest that points to the DGS server's quick_status.php
///   page.
/// - The response to the NSURLRequest is examined in order to detect whether
///   the request succeeded or failed
/// - If the request failed, the check is terminated
/// - If the request succeeded, the content of quick_status.php page is
///   examined. The games listed are extracted. The check is terminated.
/// - Before the delegate is informed that the check has finished (regardless
///   of success or failure), we do the following:
///   - Store the results of the check in the DGSMonXServer object that we have
///     been checking
///   - Remove any cookies that were generated by our own checking procedure
///   - Restore any cookies that were temporarily removed at the very beginning
///     of the checking procedure
// -----------------------------------------------------------------------------
@interface DGSMonXServerCheck : NSObject <DGSMonXServerQuickStatusDelegate>
{
@private
  DGSMonXServer* server;
  id<DGSMonXServerCheckDelegate> delegate;
  DGSMonXServerQuickStatus* quickStatus;
}

+ (DGSMonXServerCheck*) serverCheckWithServer:(DGSMonXServer*)server delegate:(id<DGSMonXServerCheckDelegate>)delegate;
- (void) cancel;

/// @brief The Dragon Go Server to perform the game state check on.
@property(nonatomic, assign) DGSMonXServer* server;
/// @brief The delegate that will be informed when the server check has
/// finished.
@property(nonatomic, assign) id<DGSMonXServerCheckDelegate> delegate;

@end
