// -----------------------------------------------------------------------------
// Copyright 2009-2012 Patrick Näf (herzbube@herzbube.ch)
//
// This file is part of the dgsmonX project.
//
// dgsmonX is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// dgsmonX is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with dgsmonX. If not, see <http://www.gnu.org/licenses/>.
// -----------------------------------------------------------------------------


// Project includes
#import "DGSMonXServerCheck.h"
#import "DGSMonXServer.h"


/// @brief This category declares private methods for the DGSMonXServerCheck
/// class.
@interface DGSMonXServerCheck()
/// @name Initialization and deallocation
//@{
- (id) init;
- (void) dealloc;
//@}
/// @name Game state check methods
//@{
- (void) check;
- (void) checkSucceeded:(int)numberOfGames;
- (void) checkFailed:(NSString*)errorDetails;
- (void) checkCancelled;
//@}
/// @name DGSMonXServerQuickStatusDelegate protocol methods
//@{
- (void) quickStatusSucceeded:(DGSMonXServerQuickStatus*)serverQuickStatus numberOfGames:(int)numberOfGames;
- (void) quickStatusFailed:(DGSMonXServerQuickStatus*)serverQuickStatus errorDetails:(NSString*)errorDetails;
- (void) quickStatusCancelled:(DGSMonXServerQuickStatus*)serverQuickStatus;
//@}
/// @name Privately declared property
//@{
@property(nonatomic, retain) DGSMonXServerQuickStatus* quickStatus;
//@}
@end


@implementation DGSMonXServerCheck

@synthesize server;
@synthesize delegate;
@synthesize quickStatus;


// -----------------------------------------------------------------------------
/// @brief Convenience constructor. Creates a DGSMonXServerCheck instance that
/// performs a game state check on @a server and calls back @a delegate when
/// the check is done.
///
/// @a server and @a delegate must not be @e nil.
// -----------------------------------------------------------------------------
+ (DGSMonXServerCheck*) serverCheckWithServer:(DGSMonXServer*)server delegate:(id<DGSMonXServerCheckDelegate>)delegate;
{
  DGSMonXServerCheck* serverCheck = [[DGSMonXServerCheck alloc] init];
  if (serverCheck)
  {
    [serverCheck autorelease];
    serverCheck.server = server;
    serverCheck.delegate = delegate;
    [serverCheck check];
  }
  return serverCheck;
}

// -----------------------------------------------------------------------------
/// @brief Initializes a DGSMonXServerCheck object.
///
/// @note This is the designated initializer of DGSMonXServerCheck.
// -----------------------------------------------------------------------------
- (id) init
{
  // Call designated initializer of superclass (NSObject)
  self = [super init];
  if (! self)
    return nil;
  
  self.server = nil;
  self.delegate = nil;
  self.quickStatus = nil;
  
  return self;
}

// -----------------------------------------------------------------------------
/// @brief Deallocates memory allocated by this DGSMonXServerCheck object.
// -----------------------------------------------------------------------------
- (void) dealloc
{
  self.quickStatus = nil;
  [super dealloc];
}

// -----------------------------------------------------------------------------
/// @brief Launches the game state check. The method returns immediately and the
/// check proceeds asynchronously, due to the asynchronous nature of
/// NSURLConnection.
// -----------------------------------------------------------------------------
- (void) check
{
  self.quickStatus = [[[DGSMonXServerQuickStatus alloc] initWithServer:self.server delegate:self] autorelease];
}

// -----------------------------------------------------------------------------
/// @brief Updates the server object on behalf of which the game state check is
/// performed, and notifies the delegate that the check succeeded.
///
/// @a numberOfGames is the number of games detected during the check for which
/// it is the user's turn.
// -----------------------------------------------------------------------------
- (void) checkSucceeded:(int)numberOfGames
{
  [self.server setCheckState:DidFinishCheckState];
  [self.server setCheckSucceeded:true];
  [self.server setNumberOfGames:numberOfGames];

  [self.delegate gameStateCheckDidFinish:self];
}

// -----------------------------------------------------------------------------
/// @brief Updates the server object on behalf of which the game state check is
/// performed, and notifies the delegate that the check failed.
///
/// @a errorDetails describes the reason for the failure.
// -----------------------------------------------------------------------------
- (void) checkFailed:(NSString*)errorDetails
{
  [self.server setCheckState:DidFinishCheckState];
  [self.server setCheckSucceeded:false];
  [self.server setErrorDetails:errorDetails];

  [self.delegate gameStateCheckDidFinish:self];
}

// -----------------------------------------------------------------------------
/// @brief Updates the server object on behalf of which the game state check is
/// performed, and notifies the delegate that the check was cancelled.
// -----------------------------------------------------------------------------
- (void) checkCancelled
{
  [self.server setCheckState:WasCancelledCheckState];

  [self.delegate gameStateCheckDidFinish:self];
}

// -----------------------------------------------------------------------------
/// @brief Terminates the game state check. Calling this method has no effect
/// if no game state check is in progress.
// -----------------------------------------------------------------------------
- (void) cancel
{
  if ([self.server checkState] != InProgressCheckState)
    return;
  [self.quickStatus cancel];
}

// -----------------------------------------------------------------------------
/// @brief DGSMonXServerQuickStatusDelegate protocol method
// -----------------------------------------------------------------------------
- (void) quickStatusSucceeded:(DGSMonXServerQuickStatus*)serverQuickStatus numberOfGames:(int)numberOfGames
{
  [self checkSucceeded:numberOfGames];
}

// -----------------------------------------------------------------------------
/// @brief DGSMonXServerQuickStatusDelegate protocol method
// -----------------------------------------------------------------------------
- (void) quickStatusFailed:(DGSMonXServerQuickStatus*)serverQuickStatus errorDetails:(NSString*)errorDetails
{
  [self checkFailed:errorDetails];
}

// -----------------------------------------------------------------------------
/// @brief DGSMonXServerQuickStatusDelegate protocol method
// -----------------------------------------------------------------------------
- (void) quickStatusCancelled:(DGSMonXServerQuickStatus*)serverQuickStatus
{
  [self checkCancelled];
}

@end
