// -----------------------------------------------------------------------------
// Copyright 2009-2012 Patrick Näf (herzbube@herzbube.ch)
//
// This file is part of the dgsmonX project.
//
// dgsmonX is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// dgsmonX is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with dgsmonX. If not, see <http://www.gnu.org/licenses/>.
// -----------------------------------------------------------------------------


// Project includes
#import "DGSMonXServerQuickStatus.h"
#import "DGSMonXServer.h"
#import "DGSMonXUtilities.h"


/// @brief This category declares private methods for the DGSMonXServerQuickStatus
/// class.
@interface DGSMonXServerQuickStatus()
/// @name Initialization and deallocation
//@{
- (void) dealloc;
//@}
/// @name NSURLConnectionDelegate protocol methods
//@{
- (void) connection:(NSURLConnection*)connection didReceiveResponse:(NSURLResponse*)response;
- (void) connection:(NSURLConnection*)connection didReceiveData:(NSData*)data;
- (void) connection:(NSURLConnection*)connection didFailWithError:(NSError*)error;
- (void) connectionDidFinishLoading:(NSURLConnection*)connection;
//@}
/// @name Helpers
//@{
- (void) retrieveQuickStatus;
- (void) quickStatusSucceeded:(int)numberOfGames;
- (void) quickStatusFailed:(NSString*)errorDetails;
- (int) processQuickStatusData:(NSMutableData*)quickStatusData withEncoding:(NSString*)encodingName errorDetails:(NSString**)errorDetails;
//@}
/// @name Privately declared properties
//@{
/// @brief The most recent connection used for an HTTP protocol exchange.
@property(nonatomic, retain) NSURLConnection* connection;
/// @brief The most recent response received during an HTTP protocol exchange.
@property(nonatomic, retain) NSURLResponse* receivedResponse;
/// @brief Container that holds data received during an HTTP protocol
/// exchange.
@property(nonatomic, retain) NSMutableData* receivedData;
//@}
@end


@implementation DGSMonXServerQuickStatus

@synthesize server;
@synthesize delegate;
@synthesize connection;
@synthesize receivedResponse;
@synthesize receivedData;


// -----------------------------------------------------------------------------
/// @brief Initializes a DGSMonXServerQuickStatus object that requests the
/// quick_status.php page from @a aServer and calls back @a aDelegate when the
/// content of the page is available.
///
/// @a aServer and @a aDelegate must not be @e nil.
///
/// @note This is the designated initializer of DGSMonXServerQuickStatus.
// -----------------------------------------------------------------------------
- (id) initWithServer:(DGSMonXServer*)aServer delegate:(id<DGSMonXServerQuickStatusDelegate>)aDelegate;
{
  // Call designated initializer of superclass (NSObject)
  self = [super init];
  if (! self)
    return nil;

  self.server = aServer;
  self.delegate = aDelegate;

  // This is assigned when the game state check begins
  self.connection = nil;
  // This is assigned/retained/released during a HTTP protocol exchange
  self.receivedResponse = nil;
  self.receivedData = [NSMutableData data];

  // Initiate retrieval of the quick status page
  [self retrieveQuickStatus];

  return self;
}

// -----------------------------------------------------------------------------
/// @brief Deallocates memory allocated by this DGSMonXServerQuickStatus object.
// -----------------------------------------------------------------------------
- (void) dealloc
{
  self.connection = nil;
  self.receivedResponse = nil;
  self.receivedData = nil;
  // Not strictly necessary, these are not retained
  self.server = nil;
  self.delegate = nil;
  [super dealloc];
}

// -----------------------------------------------------------------------------
/// @brief Initiates retrieval of the quick status page.
// -----------------------------------------------------------------------------
- (void) retrieveQuickStatus
{
  // Construct request for quick status page
  NSString* quickStatusPageWithQueryString = [NSString stringWithFormat:@"%@?version=%d&userid=%@&passwd=%@",
                                              quickStatusPage,
                                              quickStatusPageVersion,
                                              [self.server userID],
                                              [self.server userPassword]];
  NSURLRequest* quickStatusPageRequest = [DGSMonXUtilities request:self.server forPage:quickStatusPageWithQueryString];
  if (! quickStatusPageRequest)
  {
    [self quickStatusFailed:[NSString stringWithFormat:@"Failed to construct HTTP request for quick status page"]];
    return;
  }

  // Create the connection that retrieves the quick status page. The connection
  // immediately starts loading the data.
  self.connection = [[NSURLConnection alloc] initWithRequest:quickStatusPageRequest delegate:self];
  if (! self.connection)
  {
    [self quickStatusFailed:[NSString stringWithFormat:@"Connection for quick status page could not be initialized"]];
    return;
  }
}

// -----------------------------------------------------------------------------
/// @brief Sent when @a connection has received sufficient data from the server
/// to construct @a response.
///
/// This implementation does the following things:
/// - initialize a data structure that will later be used by
///   connection:didReceiveData:() to store content data that is incrementally
///   received from the server
/// - store @a response so that connectionDidFinishLoading:() can evaluate
///   which page the data came from
///
/// @note This is an NSURLConnectionDelegate protocol method.
// -----------------------------------------------------------------------------
- (void) connection:(NSURLConnection*)connection didReceiveResponse:(NSURLResponse*)response
{
  if (self.receivedResponse)
    [self.receivedResponse autorelease];
  self.receivedResponse = [response retain];
  [self.receivedData setLength:0];
}

// -----------------------------------------------------------------------------
/// @brief Sent as @a connection loads data incrementally.
///
/// This implementation appends @a data to the data structure previously
/// initialized by connection:didReceiveResponse:().
///
/// @note This is an NSURLConnectionDelegate protocol method.
// -----------------------------------------------------------------------------
- (void) connection:(NSURLConnection*)connection didReceiveData:(NSData*)data
{
  [self.receivedData appendData:data];
}

// -----------------------------------------------------------------------------
/// @brief Sent when @a connection fails to load its request successfully.
///
/// This implementaton terminates the quick status request by invoking
/// quickStatusFailed:(). The error details are taken from the message
/// encapsulated by @a error.
///
/// @note This happens for instance if there is no network connection to the
/// server (e.g. no internet connection, or the server name is wrong)
///
/// @note This is an NSURLConnectionDelegate protocol method.
// -----------------------------------------------------------------------------
- (void) connection:(NSURLConnection*)connection didFailWithError:(NSError*)error
{
  [self quickStatusFailed:[error localizedDescription]];
}

// -----------------------------------------------------------------------------
/// @brief Sent when @a connection has finished loading successfully.
///
/// This implementaton checks the NSURLResponse object that was cached by
/// connection:didReceiveResponse:() in order to determine what should be done
/// next. The NSURLResponse may refer to the following pages:
/// - quick_status.php: It is assumed that retrieval of the quick status page
///   succeeded. The retrieval process is concluded by invoking
///   quickStatusSucceeded().
/// - Anything else: It is assumed that retrieval of the quick status page
///   failed. The retrieval process is terminated by invoking
///   quickStatusFailed:(). If the URL response referred to error.php, the error
///   details are taken from the query string that has been supplied to
///   error.php.
///
/// @note This is an NSURLConnectionDelegate protocol method.
// -----------------------------------------------------------------------------
- (void) connectionDidFinishLoading:(NSURLConnection*)connection
{
  int statusCode = [(NSHTTPURLResponse*)self.receivedResponse statusCode];
  if (statusCode >= 400 && statusCode <= 599)
  {
    [self quickStatusFailed:[NSString stringWithFormat:@"HTTP status code %d", statusCode]];
    return;
  }

  NSURL* responseURL = [self.receivedResponse URL];
  NSString* pageName = (NSString*)[[[responseURL path] componentsSeparatedByString:@"/"] lastObject];
  if (NSOrderedSame == [pageName compare:quickStatusPage])
  {
    NSString* encodingName = [receivedResponse textEncodingName];
    NSString* errorDetails = nil;
    int numberOfGames = [self processQuickStatusData:receivedData withEncoding:encodingName errorDetails:&errorDetails];
    if (numberOfGames > 0)
      [self quickStatusSucceeded:numberOfGames];
    else
      [self quickStatusFailed:errorDetails];
  }
  else
  {
    NSString* errorDetails = nil;
    if (NSOrderedSame == [pageName compare:errorPage])
      errorDetails = [DGSMonXUtilities errorDetailsFromErrorPageURL:responseURL];
    if (! errorDetails)
      errorDetails = @"Retrieval of quick_status.php failed for unknown reason";
    [self quickStatusFailed:errorDetails];
  }
}

// -----------------------------------------------------------------------------
/// @brief Notifies the delegate that retrieval of the quick status page has
/// finished successfully.
// -----------------------------------------------------------------------------
- (void) quickStatusSucceeded:(int)numberOfGames
{
  [self.delegate quickStatusSucceeded:self numberOfGames:numberOfGames];
}

// -----------------------------------------------------------------------------
/// @brief Notifies the delegate that retrieval of the quick status page has
/// finished with failure.
// -----------------------------------------------------------------------------
- (void) quickStatusFailed:(NSString*)errorDetails
{
  [self.delegate quickStatusFailed:self errorDetails:errorDetails];
}

// -----------------------------------------------------------------------------
/// @brief Terminates retrieval of the quick status page and notifies the
/// delegate that retrieval has been cancelled.
// -----------------------------------------------------------------------------
- (void) cancel
{
  [self.connection cancel];   // we should not get any more delegate messages
  [self.delegate quickStatusCancelled:self];
}

// -----------------------------------------------------------------------------
/// @brief Processes quick status page data @a quickStatusData which is encoded
/// with @a encodingName.
///
/// Returns the number of games (a value >= 0) if processing is successful.
/// Returns -1 if processing failed. In this case, @a errorDetails is also set
/// to point to a descriptive error text.
///
/// @note @a errorDetails must not be nil.
///
/// The most recent source for quick_status.php is available here:
/// http://dragongoserver.cvs.sourceforge.net/viewvc/dragongoserver/DragonGoServer/quick_status.php?view=log
// -----------------------------------------------------------------------------
- (int) processQuickStatusData:(NSMutableData*)quickStatusData withEncoding:(NSString*)encodingName errorDetails:(NSString**)errorDetails
{
  /// @todo Find a function that converts encoding names (NSString*) into
  /// encoding enumeration values for us.
  NSStringEncoding encoding;
  if (! encodingName)
    encoding = NSISOLatin1StringEncoding;
  else if (NSOrderedSame == [encodingName compare:@"utf-8"])
    encoding = NSUTF8StringEncoding;
  else if (NSOrderedSame == [encodingName compare:@"iso-8859-1"])
    encoding = NSISOLatin1StringEncoding;
  else
    encoding = NSISOLatin1StringEncoding;
  NSString* dataAsString = [[NSString alloc] initWithData:quickStatusData encoding:encoding];
  if (! dataAsString)
  {
    *errorDetails = [NSString stringWithFormat:@"Unable to decode quick status page data using encoding %@", encodingName];
    return -1;
  }
  NSArray* lines = [dataAsString componentsSeparatedByString:@"\n"];
  [dataAsString release];

  int numberOfGames = 0;
  for (NSString* line in lines)
  {
    if (0 == line.length)
      continue;
    else if ([line hasPrefix:@"[#"])
    {
      // Ignore all comment lines that are not errors
      if (! [line hasPrefix:@"[#Error: "])
        continue;
      *errorDetails = [DGSMonXUtilities errorDetailsFromQuickStatusErrorLine:line];
      return -1;
    }
    else if ([line hasPrefix:@"G,"])  // games start with "G"
      numberOfGames++;
    else if ([line hasPrefix:@"MGP,"])  // multi-player games starts with "MPG"
      numberOfGames++;
    else
    {
      // Ignore everything else. Bulletins start with "B", messages start
      // with "M".
    }
  }

  return numberOfGames;
}

@end
