// -----------------------------------------------------------------------------
// Copyright 2009-2012 Patrick Näf (herzbube@herzbube.ch)
//
// This file is part of the dgsmonX project.
//
// dgsmonX is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// dgsmonX is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with dgsmonX. If not, see <http://www.gnu.org/licenses/>.
// -----------------------------------------------------------------------------


// Project includes
#import "DGSMonXUserDefaultsUpdater.h"


// -----------------------------------------------------------------------------
/// @brief User defaults keys and values that are no longer active but that
/// are still used to perform upgrades.
// -----------------------------------------------------------------------------
//@{
NSString* notifyAnimateTitleKey = @"NotifyAnimateTitle";
NSString* notifyAnimateTitleIntervalKey = @"NotifyAnimateTitleInterval";
NSString* notifyAnimateStatusItemKey = @"NotifyAnimateStatusItem";
//@}

// -----------------------------------------------------------------------------
/// @brief Class extension with private methods for DGSMonXUserDefaultsUpdater.
// -----------------------------------------------------------------------------
@interface DGSMonXUserDefaultsUpdater()
+ (void) upgradeToVersion1;
+ (void) upgradeToVersion2;
@end


@implementation DGSMonXUserDefaultsUpdater


// -----------------------------------------------------------------------------
/// @brief Performs all the required upgrades. Returns the number of upgrades
/// performed.
///
/// @retval 0 No upgrades were performed.
/// @retval >0 The number of upgrades that were performed.
/// @retval -1 A downgrade was performed.
// -----------------------------------------------------------------------------
+ (int) upgrade
{
  NSUserDefaults* userDefaults = [NSUserDefaults standardUserDefaults];
  int applicationDomainVersion = [[userDefaults valueForKey:userDefaultsVersionApplicationDomainKey] intValue];
  int registrationDomainVersion = [[userDefaults valueForKey:userDefaultsVersionRegistrationDomainKey] intValue];


  int numberOfUpgradesPerformed = 0;  // aka the return value :-)
  if (applicationDomainVersion == registrationDomainVersion)
  {
    // nothing to do
  }
  else if (applicationDomainVersion > registrationDomainVersion)
  {
    // TODO perform downgrade
    numberOfUpgradesPerformed = -1;
  }
  else
  {
    while (applicationDomainVersion < registrationDomainVersion)
    {
      // Incrementally perform upgrades. We allow for gaps in the user defaults
      // versioning scheme.
      ++applicationDomainVersion;
      NSString* upgradeMethodName = [NSString stringWithFormat:@"upgradeToVersion%d", applicationDomainVersion];
      SEL upgradeSelector = NSSelectorFromString(upgradeMethodName);
      if ([[DGSMonXUserDefaultsUpdater class] respondsToSelector:upgradeSelector])
      {
        // TODO How do we learn of success/failure of upgradeSelector, and how
        // do we react to failure?
        [[DGSMonXUserDefaultsUpdater class] performSelector:upgradeSelector];
        ++numberOfUpgradesPerformed;
        // Update the application domain version number
        [userDefaults setValue:[NSNumber numberWithInt:applicationDomainVersion]
                        forKey:userDefaultsVersionApplicationDomainKey];
      }
    }
  }

  // Perform final check if the cumulative effect of all upgrades had the
  // desired effect.
  int realApplicationDomainVersion = [[userDefaults valueForKey:userDefaultsVersionApplicationDomainKey] intValue];
  if (realApplicationDomainVersion != registrationDomainVersion)
  {
    // TODO: Display an alert that notifies the user of the problem. The alert
    // should probably recommend a re-install. Also decide whether to allow the
    // user to continue, or to gracefully shutdown the application.
  }

  return numberOfUpgradesPerformed;
}

// -----------------------------------------------------------------------------
/// @brief Performs the incremental upgrade to the user defaults format
/// version 1.
// -----------------------------------------------------------------------------
+ (void) upgradeToVersion1
{
  NSUserDefaults* userDefaults = [NSUserDefaults standardUserDefaults];

  // Rename notifyAnimateTitleKey to notifyAnimateStatusItemKey
  id object = [userDefaults objectForKey:notifyAnimateTitleKey];
  if (object)  // is nil if the key is not present
  {
    BOOL animateStatusItem = [userDefaults boolForKey:notifyAnimateTitleKey];
    [userDefaults setBool:animateStatusItem forKey:notifyAnimateStatusItemKey];
    [userDefaults removeObjectForKey:notifyAnimateTitleKey];
  }

  // Rename notifyAnimateTitleIntervalKey to notifyAnimateStatusItemIntervalKey
  object = [userDefaults objectForKey:notifyAnimateTitleIntervalKey];
  if (object)  // is nil if the key is not present
  {
    float animateInterval = [userDefaults floatForKey:notifyAnimateTitleIntervalKey];
    [userDefaults setFloat:animateInterval forKey:notifyAnimateStatusItemIntervalKey];
    [userDefaults removeObjectForKey:notifyAnimateTitleIntervalKey];
  }
}

// -----------------------------------------------------------------------------
/// @brief Performs the incremental upgrade to the user defaults format
/// version 2.
// -----------------------------------------------------------------------------
+ (void) upgradeToVersion2
{
  NSUserDefaults* userDefaults = [NSUserDefaults standardUserDefaults];
  
  // Rename notifyAnimateStatusItemKey to notifyStyleKey and transform value
  // from data type boolean to integer
  id object = [userDefaults objectForKey:notifyAnimateStatusItemKey];
  if (object)  // is nil if the key is not present
  {
    BOOL animateStatusItem = [userDefaults boolForKey:notifyAnimateStatusItemKey];
    enum DGSMonXNotificationStyle notificationStyle;
    if (animateStatusItem)
      notificationStyle = NotificationStyleAnimation;
    else
      notificationStyle = NotificationStyleStatic;
    [userDefaults setInteger:notificationStyle forKey:notifyStyleKey];
    [userDefaults removeObjectForKey:notifyAnimateStatusItemKey];
  }
}

@end
