// -----------------------------------------------------------------------------
// Copyright 2011-2012 Patrick Näf (herzbube@herzbube.ch)
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// -----------------------------------------------------------------------------


// Project includes
#import "DGSMonXUtilities.h"
#import "DGSMonXServer.h"


@implementation DGSMonXUtilities

// -----------------------------------------------------------------------------
/// @brief Factory method that creates and returns an autoreleased NSURLRequest
/// object that refers to the page named @a pageName located on the Dragon Go
/// Server @a server.
///
/// Returns @e nil if any error occurs.
// -----------------------------------------------------------------------------
+ (NSURLRequest*) request:(DGSMonXServer*)server forPage:(NSString*)pageName;
{
  NSString* baseURL = [server baseURL];
  NSString* pageURL = [baseURL stringByAppendingString:pageName];
  NSURLRequest* request = [NSURLRequest requestWithURL:[NSURL URLWithString:pageURL]];
  return [[request retain] autorelease];
}

// -----------------------------------------------------------------------------
/// @brief Extracts and returns error details from the query string that has
/// been supplied to the error page URL. Returns nil if no error details are
/// available.
///
/// For instance, an URL might look like this:
/// http://www.dragongoserver.net/error.php?err=not_logged_in&debugmsg=status&eid=113984
/// The result of invoking this method for this URL would be "not_logged_in".
// -----------------------------------------------------------------------------
+ (NSString*) errorDetailsFromErrorPageURL:(NSURL*)responseURL
{
  NSString* queryString = [responseURL query];
  NSArray* fields = [queryString componentsSeparatedByString:@"&"];
  for (NSString* field in fields)
  {
    NSArray* fieldNameAndValue = [field componentsSeparatedByString:@"="];
    if (! fieldNameAndValue || [fieldNameAndValue count] != 2)
      continue;
    NSString* fieldName = [fieldNameAndValue objectAtIndex:0];
    if (! [fieldName isEqualToString:@"err"])
      continue;
    return [fieldNameAndValue lastObject];
  }
  return nil;
}

// -----------------------------------------------------------------------------
/// @brief Extracts and returns error details from an error comment line printed
/// by quick_status.php.
///
/// For instance, an error line might look like this:
/// [#Error: wrong_password; quick_status.check_password(userid)]
/// The result of invoking this method for this URL would be "wrong_password".
// -----------------------------------------------------------------------------
+ (NSString*) errorDetailsFromQuickStatusErrorLine:(NSString*)errorLine
{
  NSString* errorDetails = [errorLine stringByReplacingOccurrencesOfString:@"[#Error: " withString:@""];
  return [[errorDetails componentsSeparatedByString:@";"] objectAtIndex:0];
}

// -----------------------------------------------------------------------------
/// @brief Prints all details of @a url to NSLog.
///
/// @note This is an internal helper used for debugging purposes.
// -----------------------------------------------------------------------------
+ (void) logURLDetails:(NSURL*)url
{
  NSStringEncoding encoding;
  NSLog(@"absoluteString = %@", [url absoluteString]);
  NSLog(@"absoluteURL = %@", [NSString stringWithContentsOfURL:[url absoluteURL] usedEncoding:&encoding error:NULL]);
  NSLog(@"baseURL = %@", [NSString stringWithContentsOfURL:[url baseURL] usedEncoding:&encoding error:NULL]);
  NSLog(@"fragment = %@", [url fragment]);
  NSLog(@"host = %@", [url host]);
  NSLog(@"parameterString = %@", [url parameterString]);
  NSLog(@"password = %@", [url password]);
  NSLog(@"path = %@", [url path]);
  NSLog(@"port = %@", [url port]);
  NSLog(@"query = %@", [url query]);
  NSLog(@"relativePath = %@", [url relativePath]);
  NSLog(@"relativeString = %@", [url relativeString]);
  NSLog(@"resourceSpecifier = %@", [url resourceSpecifier]);
  NSLog(@"scheme = %@", [url scheme]);
  NSLog(@"standardizedURL = %@", [NSString stringWithContentsOfURL:[url standardizedURL] usedEncoding:&encoding error:NULL]);
  NSLog(@"user = %@", [url user]);
}

@end
